*==============================================================================
* Program:			TaskScheduler.PRG
* Purpose:			Provides class to create Windows Task Scheduler tasks
* Author:			Doug Hennig, Stonefield Software Inc.
* Last Revision:	09/25/2017
* Web site:			http://www.stonefieldquery.com
* Notes:			see comments for each property listed below
*					see Test.PRG for examples
*					for Windows XP or earlier, use REGSVR32 on
*						TaskScheduler.DLL to register it
*==============================================================================

#define ccCR chr(13)

define class TaskScheduler as Custom
	ErrorMessage = ''
		&& the text of any error message
	ErrorReason = ''
		&& the reason for the error
	ErrorCode = 0
		&& an error code; can be used to localize error messages
		&& 1 = scheduler class not found
		&& 2 = scheduler failed to instantiate
		&& 3 = a new task could not be created
		&& 4 = the task properties could not be set
		&& 5 = the scheduler settings could not be saved
		&& 6 = the task could not be registered

* Schedule settings.

	TaskName = ''
		&& the name of the task
	Description = ''
		&& the description of the task
	ScheduleType = 0
		&& the type of schedule: 1 = daily, 2 = weekly, 3 = monthly,
		&& 4 = monthly day-of-week
	StartTime = {/:}
		&& the starting date and time for the task
	Interval = 1
		&& the frequency for a daily or weekly schedule
	EXEName = ''
		&& the name and path of the program to run
	EXEParameters = ''
		&& parameters to pass to the program
	AuthorName = ''
		&& the name of the task author
	UserName = ''
		&& the name of the user to log in as
	Password = ''
		&& the password for the user
	StartWhenAvailable = .T.
		&& .T. to start the task as soon as possible after a scheduled start is
		&& missed; not supported in all versions (only Vista and above)
	FolderName = ''
		&& the name of the folder the task goes in; currently unsupported
	RunOnLast = .F.
		&& .T. to run on the last week or last day of the month; not supported
		&& in all versions (only Vista and above)
	dimension DaysOfWeek[7]
		&& set the appropriate element to .T. to execute the task on that day
		&& for a weekly task; Sunday is day 1 and Saturday is day 7
	dimension DaysOfMonth[31]
		&& set the appropriate element to .T. to execute the task on that day
		&& for a monthly task
	dimension MonthsOfYear[12]
		&& set the appropriate element to .T. to execute the task in that month
		&& for a monthly task
	dimension WeeksOfMonth[4]
		&& set the appropriate element to .T. to execute the task in that week
		&& for a monthly day-of-week task

* Protected properties.

	protected oScheduler
	oScheduler = .NULL.
		&& a reference to the Task Scheduler object
	protected cSchedulerClass
	cSchedulerClass = ''
		&& the class for the scheduler object

	procedure Init
		with This
			try

* Instantiate and set up the scheduler.

				.oScheduler = createobject(.cSchedulerClass)
				.SetupScheduler()

* Handle the scheduler DLL not installed.

			catch to loException when loException.ErrorNo = 1733
				.ErrorMessage = 'The scheduler class cannot be located.'
				.ErrorReason  = loException.Message
				.ErrorCode    = 1

* Handle any other errors.

			catch to loException
				.ErrorMessage = 'The scheduler failed.' + ccCR + ccCR + ;
					loException.Message
				.ErrorReason  = loException.Message
				.ErrorCode    = 2
			endtry
		endwith
	endproc

	procedure Destroy
		This.oScheduler = .NULL.
	endproc

	procedure CreateTask
		local loTask, ;
			llReturn
		with This

* Get a reference to a task object.

			loTask = .NewTask()

* Set the task settings, set up the trigger, and register the task.

			if vartype(loTask) = 'O'
				llReturn = .SetSettings(loTask)
				llReturn = llReturn and .SetTrigger(loTask)
				llReturn = llReturn and .RegisterTask(loTask)
			endif vartype(loTask) = 'O'
		endwith
		return llReturn
	endproc

* Convert logical elements in an array into a bitmap.

	protected procedure CreateBitMapFromArray(taArray)
		local lnBitmap, ;
			lnI
		lnBitmap = 0
		if .F.
			dimension taArray[1]	&& fool the project manager into not giving
									&& an error
		endif .F.
		for lnI = 1 to alen(taArray)
			lnBitmap = iif(taArray[lnI], bitset(lnBitMap, lnI - 1), lnBitmap)
		next lnI
		return lnBitmap
	endproc

* Convert a bitmap into logical elements in an array.

	protected procedure CreateArrayFromBitMap(tnBitmap, taArray)
		for lnI = 1 to alen(taArray)
			taArray[lnI] = bittest(tnBitMap, lnI - 1)
		next lnI
	endproc

* Abstract methods, implemented in subclasses.

	protected procedure SetupScheduler
	endproc

	protected procedure NewTask
	endproc

	protected procedure SetTrigger(toTask)
	endproc

	protected procedure RegisterTask(toTask)
	endproc
enddefine

*********************************************
define class XPTaskScheduler as TaskScheduler
*********************************************
	cSchedulerClass = 'Scheduler.SchAgent'

	protected procedure SetupScheduler
		This.oScheduler.Refresh()
	endproc

	protected procedure NewTask
		local lcTask, ;
			lcEXE, ;
			lcParameters, ;
			loCurrTask, ;
			loTask, ;
			loPossible, ;
			loException as Exception
		This.ErrorMessage = ''
		try

* Go through the existing tasks and see if the one we want to create already
* exists. If we can't find it by name, try to match on the command line.

			lcTask       = upper(This.TaskName)
			lcEXE        = upper(This.EXEName)
			lcParameters = upper(This.EXEParameters)
			for each loCurrTask in This.oScheduler
				do case
					case upper(loCurrTask.Name) == lcTask
						loTask = loCurrTask
						exit
					case upper(loCurrTask.ApplicationName) == lcEXE and ;
						upper(loCurrTask.CommandLine) == lcParameters
						loPossible = loTask
				endcase
			next loTask

* If we didn't find the specified task but have another one with the same
* application and parameters, use it instead.

			if vartype(loTask) <> 'O' and vartype(loPossible) = 'O'
				loTask = loPossible
			endif vartype(loTask) <> 'O' ...

* If we didn't find a match, create a new task

			if vartype(loTask) <> 'O'
				loTask = This.oScheduler.CreateTask(This.TaskName)
			endif vartype(loTask) <> 'O'

* Something went wrong.

		catch to loException
			This.ErrorMessage = 'A new task could not be created.' + ccCR + ;
				ccCR + loException.Message
			This.ErrorReason  = loException.Message
			This.ErrorCode    = 3
			loTask = .NULL.
		endtry
		return loTask
	endproc

	protected procedure SetSettings(toTask)
		local llReturn, ;
			loException as Exception
		This.ErrorMessage = ''
		try
			with toTask
				.ApplicationName  = This.EXEName
				.CommandLine      = This.EXEParameters
				.Comment          = This.Description
				.WorkingDirectory = '"' + addbs(justpath(This.EXEName)) + '"'
				.Creator          = This.AuthorName
				.SetAccountInformation(This.UserName, This.Password)
			endwith
			llReturn = .T.

* Something went wrong.

		catch to loException
			This.ErrorMessage = 'The task properties could not be set.' + ;
				ccCR + ccCR + loException.Message
			This.ErrorReason  = loException.Message
			This.ErrorCode    = 4
		endtry
		return llReturn
	endproc

	protected procedure SetTrigger(toTask)
		local loTrigger, ;
			lnType, ;
			lnHour, ;
			lcAMPM, ;
			laArray[1], ;
			lnI, ;
			llReturn, ;
			loException as Exception
		This.ErrorMessage = ''
		try

* If we alreasy have a trigger for the task, get it so we can overwrite its
* settings. Otherwise, create a new one.

			if toTask.Triggers.Count > 0
				loTrigger = toTask.Triggers.Item(1)
			else
				loTrigger = toTask.Triggers.Add()
			endif vartype(.oTask) = 'O' ...
			lnType = This.ScheduleType

* Set the trigger properties.

			with loTrigger
				lnHour = hour(This.StartTime)
				do case
					case lnHour = 12
						lcAMPM = ' PM'
					case lnHour > 12
						lnHour = lnHour - 12
						lcAMPM = ' PM'
					otherwise
						lcAMPM = ' AM'
				endcase
				.StartTime   = transform(lnHour) + ':' + ;
					transform(minute(This.StartTime)) + lcAMPM
				.TriggerType = lnType
				do case

* Daily task.

					case lnType = 1
						.BeginDay     = ttod(This.StartTime)
						.DaysInterval = This.Interval

* Weekly task.

					case lnType = 2
						.WeeksInterval = This.Interval
						acopy(This.DaysOfWeek, laArray)
						.DaysOfTheWeek = This.CreateBitMapFromArray(@laArray)

* Monthly task.

					case lnType = 3
						for lnI = 1 to alen(This.DaysOfMonth)
							if This.DaysOfMonth[lnI]
								.Day = lnI
								exit
							endif This.DaysOfMonth[lnI]
						next lnI
						acopy(This.MonthsOfYear, laArray)
						.Months = This.CreateBitMapFromArray(@laArray)

* Monthly day-of-week task.

					case lnType = 4
						acopy(This.DaysOfWeek, laArray)
						.DaysOfTheWeek = This.CreateBitMapFromArray(@laArray)
						if This.RunOnLast
							.Week = 5
						else
							for lnI = 1 to alen(This.WeeksOfMonth)
								if This.WeeksOfMonth[lnI]
									.Week = lnI
									exit
								endif This.WeeksOfMonth[lnI]
							next lnI
						endif This.RunOnLast
						dimension laArray[12]
						acopy(This.MonthsOfYear, laArray)
						.Months = This.CreateBitMapFromArray(@laArray)
				endcase
				.Update()
			endwith

* Save the schedule.

			toTask.Triggers.Update()
			llReturn = .T.

* Something went wrong.

		catch to loException
			This.ErrorMessage = 'The scheduler settings could not be ' + ;
				'saved.' + ccCR + ccCR + loException.Message
			This.ErrorReason  = loException.Message
			This.ErrorCode    = 5
		endtry
		return llReturn
	endproc

	protected procedure RegisterTask(toTask)
		local llReturn, ;
			loException as Exception
		This.ErrorMessage = ''
		try
			toTask.Save()
			This.oScheduler.Refresh()
			llReturn = .T.

* Something went wrong.

		catch to loException
			This.ErrorMessage = 'The task could not be registered.' + ;
				ccCR + ccCR + loException.Message
			This.ErrorReason  = loException.Message
			This.ErrorCode    = 6
		endtry
		return llReturn
	endproc
enddefine

************************************************
define class VistaTaskScheduler as TaskScheduler
************************************************

* Vista scheduler enums.

	#define TASK_ACTION_EXEC 0
	#define TASK_TRIGGER_DAILY 2
	#define TASK_TRIGGER_WEEKLY 3
	#define TASK_TRIGGER_MONTHLY 4
	#define TASK_TRIGGER_MONTHLYDOW 5
	#define TASK_LOGON_PASSWORD 1
	#define TASK_LOGON_INTERACTIVE_TOKEN 3
	#define TASK_LOGON_INTERACTIVE_TOKEN_OR_PASSWORD 6
	#define TASK_CREATE_OR_UPDATE 6

* Properties.

	cSchedulerClass = 'Schedule.Service'
	protected oFolder
	oFolder = .NULL.
		&& a reference to a folder object

	procedure Destroy
		This.oFolder = .NULL.
		dodefault()
	endproc
		
	protected procedure SetupScheduler
		This.oScheduler.Connect()
	endproc

	procedure GetTasks(taTasks)
		lnTasks = 0
		This.ErrorMessage = ''
		try
			loFolder = This.GetFolder(This.FolderName)
			loTasks  = loFolder.GetTasks(0)
			lnTasks  = loTasks.Count
			dimension taTasks[lnTasks, 2]
			for lnI = 1 to lnTasks
				loTask          = loTasks.Item(lnI)
				taTasks[lnI, 1] = loTask.Name
				taTasks[lnI, 2] = loTask
			next loTask
		catch to loException
			This.ErrorMessage = 'Tasks could not be enumerated.' + ccCR + ;
				ccCR + loException.Message
			This.ErrorReason  = loException.Message
		endtry
		return lnTasks
	endproc

	procedure GetTask(tcName)
		loTask = .NULL.
		This.ErrorMessage = ''
		try
			loFolder = This.GetFolder(This.FolderName)
			loTask   = loFolder.GetTask(tcName)
			This.TaskName           = tcName
			This.StartWhenAvailable = loTask.Definition.Settings.StartWhenAvailable

			loAction = loTask.Definition.Actions(1)
			with loAction
				if .Type = TASK_ACTION_EXEC
					This.EXEName       = .Path
					This.EXEParameters = .Arguments
				endif .Type = TASK_ACTION_EXEC
			endwith

			loReg = loTask.Definition.RegistrationInfo
			with loReg
				This.Description = .Description
				This.AuthorName  = .Author
			endwith

			loTrigger = loTask.Definition.Triggers(1)
			with loTrigger
				This.StartTime = ctot(left(.StartBoundary, 19))
				do case
					case .Type = TASK_TRIGGER_DAILY
						This.Interval     = .DaysInterval
						This.ScheduleType = 1
					case .Type = TASK_TRIGGER_WEEKLY
						dimension laArray[alen(This.DaysOfWeek)]
						This.CreateArrayFromBitMap(.DaysOfWeek, @laArray)
						acopy(laArray, This.DaysOfWeek)

						This.Interval     = .WeeksInterval
						This.ScheduleType = 2
					case .Type = TASK_TRIGGER_MONTHLY
						dimension laArray[alen(This.DaysOfMonth)]
						This.CreateArrayFromBitMap(.DaysOfMonth, @laArray)
						acopy(laArray, This.DaysOfMonth)

						dimension laArray[alen(This.MonthsOfYear)]
						This.CreateArrayFromBitMap(.MonthsOfYear, @laArray)
						acopy(laArray, This.MonthsOfYear)

						This.RunOnLast    = .RunOnLastDayOfMonth
						This.ScheduleType = 3
					case .Type = TASK_TRIGGER_MONTHLYDOW
						dimension laArray[alen(This.DaysOfWeek)]
						This.CreateArrayFromBitMap(.DaysOfWeek, @laArray)
						acopy(laArray, This.DaysOfWeek)

						dimension laArray[alen(This.MonthsOfYear)]
						This.CreateArrayFromBitMap(.MonthsOfYear, @laArray)
						acopy(laArray, This.MonthsOfYear)

						dimension laArray[alen(This.WeeksOfMonth)]
						This.CreateArrayFromBitMap(.WeeksOfMonth, @laArray)
						acopy(laArray, This.WeeksOfMonth)

						This.RunOnLast    = .RunOnLastWeekOfMonth
						This.ScheduleType = 4
				endcase
			endwith

		catch to loException
			This.ErrorMessage = 'Task could not be retrieved.' + ccCR + ;
				ccCR + loException.Message
			This.ErrorReason  = loException.Message
		endtry
		return loTask
	endproc

	procedure DeleteTask(tcName)
		llReturn = .F.
		This.ErrorMessage = ''
		try
			loFolder = This.GetFolder(This.FolderName)
			loFolder.DeleteTask(tcName, 0)
			llReturn = .T.
		catch to loException
			This.ErrorMessage = 'Task could not be deleted.' + ccCR + ;
				ccCR + loException.Message
			This.ErrorReason  = loException.Message
		endtry
		return llReturn
	endproc

	protected procedure NewTask
		local loTask, ;
			loException as Exception
		This.ErrorMessage = ''
		try
			with This
				.oFolder = .GetFolder(.FolderName)
				loTask   = .oScheduler.NewTask(0)
			endwith

* Something went wrong.

		catch to loException
			This.ErrorMessage = 'A new task could not be created.' + ccCR + ;
				ccCR + loException.Message
			This.ErrorReason  = loException.Message
			This.ErrorCode    = 3
			loTask = .NULL.
		endtry
		return loTask
	endproc

	protected procedure GetFolder(tcFolder)
		local loFolder
*** Problem: CreateFolder needs a security descriptor and I don't know how to
*** create one. So, we'll ignore tcFolder and use the root folder for now.
		loFolder = This.oScheduler.GetFolder('\')
*		try
*			loFolder = This.oScheduler.GetFolder(tcFolder)
*		catch
*			loRootFolder = This.oScheduler.GetFolder('\')
*			loFolder = loRootFolder.CreateFolder(tcFolder, SecurityDescriptor)
*		endtry
		return loFolder
	endproc

	protected procedure SetSettings(toTask)
		local loSettings, ;
			loAction, ;
			loReg, ;
			llReturn, ;
			loException as Exception
		This.ErrorMessage = ''
		try
			loSettings = toTask.Settings
			with loSettings
				.StartWhenAvailable = This.StartWhenAvailable
				.Enabled            = .T.
				.Hidden             = .F.
			endwith

* Set the action settings.

			loAction = toTask.Actions.Create(TASK_ACTION_EXEC)
			with loAction
				.Path             = This.EXEName
				.Arguments        = This.EXEParameters
				.WorkingDirectory = addbs(justpath(This.EXEName))
			endwith

* Set the registration information.

			loReg = toTask.RegistrationInfo
			with loReg
				.Date        = ttoc(datetime(), 3)
				.Description = This.Description
				.Author      = This.AuthorName
			endwith
			llReturn = .T.

* Something went wrong.

		catch to loException
			This.ErrorMessage = 'The task properties could not be set.' + ;
				ccCR + ccCR + loException.Message
			This.ErrorReason  = loException.Message
			This.ErrorCode    = 5
		endtry
		return llReturn
	endproc

	protected procedure SetTrigger(toTask)
		local lnType, ;
			loTrigger, ;
			laArray[1], ;
			llReturn, ;
			loException as Exception
		This.ErrorMessage = ''
		try
			lnType    = This.ScheduleType
			loTrigger = toTask.Triggers.Create(lnType + 1)
			with loTrigger
				do case

* Daily task.

					case lnType = 1
						.DaysInterval = This.Interval

* Weekly task.

					case lnType = 2
						acopy(This.DaysOfWeek, laArray)
						.DaysOfWeek    = This.CreateBitMapFromArray(@laArray)
						.WeeksInterval = This.Interval

* Monthly task.

					case lnType = 3
						acopy(This.DaysOfMonth, laArray)
						.DaysOfMonth = This.CreateBitMapFromArray(@laArray)
						dimension laArray[12]
						acopy(This.MonthsOfYear, laArray)
						.MonthsOfYear        = This.CreateBitMapFromArray(@laArray)
						.RunOnLastDayOfMonth = This.RunOnLast

* Monthly day-of-week task.

					case lnType = 4
						acopy(This.DaysOfWeek, laArray)
						.DaysOfWeek = This.CreateBitMapFromArray(@laArray)
						dimension laArray[12]
						acopy(This.MonthsOfYear, laArray)
						.MonthsOfYear = This.CreateBitMapFromArray(@laArray)
						dimension laArray[4]
						acopy(This.WeeksOfMonth, laArray)
						.WeeksOfMonth         = This.CreateBitMapFromArray(@laArray)
						.RunOnLastWeekOfMonth = This.RunOnLast
				endcase
				loTimeZone = createobject('Empty')
				addproperty(loTimeZone, 'TimeZoneDesc')
				addproperty(loTimeZone, 'TimeZoneOffset')
				GetTimeZoneInfo(loTimeZone)
				lnOffset       = loTimeZone.TimeZoneOffset/3600 * 100
					&& Convert seconds to hours, then multiply by 100 so the
					&& format 99:99 works
				lcOffset       = iif(lnOffset < 0, '+', '-') + ;
					transform(abs(lnOffset), '@L 99:99')
					&& Note the sign is reversed because TimeZoneOffset is what
					&& to add to local to get UTC
				.StartBoundary = ttoc(This.StartTime, 3) + lcOffset
			endwith
			llReturn = .T.

* Something went wrong.

		catch to loException
			This.ErrorMessage = 'The scheduler settings could not be ' + ;
				'saved.' + ccCR + ccCR + loException.Message
			This.ErrorReason  = loException.Message
			This.ErrorCode    = 5
		endtry
		return llReturn
	endproc

	protected procedure RegisterTask(toTask)
		local llReturn, ;
			loException as Exception
		This.ErrorMessage = ''
		try
			This.oFolder.RegisterTaskDefinition(This.TaskName, toTask, ;
				TASK_CREATE_OR_UPDATE, This.UserName, This.Password, ;
				TASK_LOGON_PASSWORD, NULL)
			llReturn = .T.

* Something went wrong.

		catch to loException
			This.ErrorMessage = 'The task could not be registered.' + ;
				ccCR + ccCR + loException.Message
			This.ErrorReason  = loException.Message
			This.ErrorCode    = 6
		endtry
		return llReturn
	endproc
enddefine

function GetTimeZoneInfo(toObject)
local lcTimeZone, ;
	lnID, ;
	lnStandardOffset, ;
	lnDaylightOffset

* Declare the time zone information API function and get the time zone
* information.
 
#define TIME_ZONE_SIZE  172
declare integer GetTimeZoneInformation in kernel32 ;
	string @lpTimeZoneInformation
lcTimeZone = replicate(chr(0), TIME_ZONE_SIZE)
lnID       = GetTimeZoneInformation(@lcTimeZone)

* Determine the standard and daylight time offset.
 
lnStandardOffset = ctobin(substr(lcTimeZone,   1, 4), '4RS')
lnDaylightOffset = ctobin(substr(lcTimeZone, 169, 4), '4RS')

* Determine the total offset based on whether the computer is on daylight time
* or not. Get the description for the time zone.
 
if lnID = 2  && daylight time
	toObject.TimeZoneDesc   = strtran(strconv(substr(lcTimeZone, 89, 64), ;
		6), chr(0), '')
	toObject.TimeZoneOffset = (lnStandardOffset + lnDaylightOffset) * 60
else   && standard time
	toObject.TimeZoneDesc   = strtran(strconv(substr(lcTimeZone,  5, 64), ;
		6), chr(0), '')
	toObject.TimeZoneOffset = lnStandardOffset * 60
endif lnID = 2
